#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#           Mathlib 2 (sympy powered)         #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

## Cette librairie contient des fonctions mathmatiques  usage interne

import math
import cmath
import pylab
import sympy
from pylib import *

## Fonctions rapides destines essentiellement  golib
    
def produit_scalaire(u, v):
    u"""Calcule le produit scalaire des vecteurs u et v.
    u et v doivent tre de type list, tuple, array, ou geolib.Vecteur."""
    return sum(i*j for i, j in zip(u ,v))
    
    
    
def angle_vectoriel(u, v):
    u"""Renvoie une mesure sur ]-pi;pi] de l'angle form par les vecteurs u et v.
    u et v doivent tre de type list, tuple, array, ou geolib.Vecteur, et de dimension 2."""
    return cmath.log(complex(*v)/complex(*u)).imag

#~ def _angle_vectoriel_bis_(u, v):
    #~ u"Autre implmentation du produit vectoriel (un peu plus lente)."
    #~ return (cmp(u[0]*v[1]-v[0]*u[1], 0) or 1)*math.acos(produit_scalaire(u, v)/(math.hypot(*u)*math.hypot(*v)))

#~ def _angle_vectoriel_ter_(u, v):
    #~ u"Autre implmentation du produit vectoriel (plus lente)."
    #~ return angle(complex(*v)/complex(*u))


def norm2(couple):
    u"Implmentation rapide de la norme d'un vecteur."
    return math.hypot(*couple)

def norm2diff(xy1, xy2):
    u"Implmentation rapide de la norme d'un vecteur, donn par ses deux extrmits."
    return math.hypot(xy2[0] - xy1[0], xy2[1] - xy1[1])

## Fonctions de formatage (pour l'affichage entre autres)

#~ def old_nchiffres(x, n = 1):
    #~ u"Arrondi x en fournissant n chiffres significatifs. Ex: nchiffres(2345, 2)."
    
    #~ return x and round(x/10**floor(log10(abs(x))-n+1))*10**floor(log10(abs(x))-n+1) # Attention au cas x = 0 !
    
def nchiffres(x, n = 1):
    u"Arrondi x en fournissant n chiffres significatifs. Ex: nchiffres(2345, 2)."
    if x:
        k = 10**math.floor(math.log10(abs(x))-n+1)
        return round(x/k)*k
    return x # Attention au cas x = 0 !


def arrondir(x):
    u"""Arrondi le nombre : un seul chiffre significatif, compris entre 1 et 5.
    Transforme automatiquement le nombre en entier selon le cas."""
    n = nchiffres(x, math.sqrt(.5))
    if int(n) == n:
        n = int(n)
    return n


def array_zip(*args):
    return (pylab.array(item) for item in zip(*args))


# Les deux fonctions qui suivent sont bien plus rapides que la classe Polynome pour fournir des solutions approchees.
# Elles servent comme routine dans geolib par exemple.

def delta(a, b, c):
    return b**2 - 4*a*c
    
def racines(a, b, c):
    d = delta(a, b, c)
    if not d:
        return [-b/(2*a)]
    if d > 0:
        rac = math.sqrt(d)
        return [(-rac-b)/(2*a) , (rac-b)/(2*a)]
    return []




def point_dans_polygone(couple, polygone):
    u"""Dtermine si un point est  l'intrieur d'un polygone.
    D'aprs un algorithme de Paul Bourke."""
    x, y = couple
    dedans = False
    n = len(polygone)
    p1x, p1y = polygone[0]
    for i in range(1, n + 1):
        p2x, p2y = polygone[i % n]
        if min(p1y, p2y) < y <= max(p1y, p2y) and x <= max(p1x, p2x):
            if p1y != p2y:
                xinters = (y - p1y)*(p2x - p1x)//(p2y - p1y) + p1x
                if p1x == p2x or x <= xinters:
                    dedans = not dedans
        p1x, p1y = p2x, p2y

    return dedans


def poly_factor(polynome, variable, corps = None):
    u"""Factorise un polynome  une variable.
    
    Le corps peut-tre R ou C.
    Par dfaut, le corps de factorisation est celui des coefficients."""
    coeffs = polynome.as_poly().coeffs
    if corps == "R":
        if not all(coeff.is_real for coeff in coeffs):
            raise "ValueError", "factorisation dans 'R' impossible."
    elif corps is None:
        if all(coeff.is_real for coeff in coeffs):
            corps = "R"
        else:
            corps = "C"
    racines = sympy.rootfinding.roots(polynome, variable, cubics=True, quartics=True)
    coefficient = coeffs[0]
    produit = 1
    if corps == "R":
        racines_en_stock = []
        for racine, multiplicite in racines.iteritems():
            if racine.is_real:
                produit *= (variable - racine)**multiplicite
            else:
                conjuguee = racine.conjugate()
                if conjuguee in racines_en_stock:
                    produit *= (sympy.expand((variable - racine)*(variable - conjuguee)))**multiplicite
                    racines_en_stock.remove(conjuguee)
                else:
                    racines_en_stock.append(racine)
    else:
        for racine, multiplicite in racines.iteritems():
            produit *= (variable - racine)**multiplicite
    quotient, reste = sympy.div(polynome, coefficient*produit, variable)
    assert reste == 0
    return coefficient*produit*quotient
    