# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                Intervalles                  #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

#~ from objets import *
import sympy
oo = sympy.oo





class Union(object):
    u"""Une union finie d'intervalles rels.
    Lors de l'initialisation, elle est reecrite sous forme d'union disjointe et ordonne.
    
    Les opration sur les unions d'intervalles sont :
    - l'union : symbole "A+B"
    - l'intersection : symbole "A*B"
    - la diffrence : symbole "A-B"
    - le complmentaire : symbole -A
    
    Note: l'inclusion et l'appartenance sont notes de manire identique:
    "y in A"
    "A in B".
    """
    def __init__(self, *intervalles):
        self.intervalles = [i.__copy__() for i in intervalles] # important (ne pas travailler sur les originaux) !
        self.simplifier()
        if not self.intervalles:
            self.__class__ = Vide
        if len(self.intervalles) == 1:
            self.__class__ = Intervalle
            self.inf = self.intervalles[0].inf
            self.sup = self.intervalles[0].sup
            self.inf_inclus = self.intervalles[0].inf_inclus
            self.sup_inclus = self.intervalles[0].sup_inclus

    def __add__(self, y):
        "union"
        return Union(*(self.intervalles + y.intervalles))
        
    def __radd__(self, y):
        return self.__add__(y)
        
    def __mul__(self, y):
        "intersection"
        if y == Vide():
            return y
        else:
            return Union(*[i*j for i in self.intervalles for j in y.intervalles])
    
    def __pow__(self, y): # alias pour l'intersection (le symbole '^' rappelle vagument le symbole 'inter')
        return self.__add__(y)
    
    
    def __rmul__(self, y):
        self.__mul__(y)
        
    def __rpow__(self, y):
        self.__radd__(y)
    
    def simplifier(self):
        ints = [intervalle for intervalle in self.intervalles if intervalle <> Vide()]
        ints.sort()
        #print ints
        #print self.intervalles
        for i in range(len(ints) - 1):
            if ints[i].sup > ints[i + 1].inf: # les intervalles se coupent
                if ints[i + 1].inf == ints[i].inf:
                    ints[i + 1].inf_inclus = ints[i + 1].inf_inclus or ints[i].inf_inclus
                else:
                    ints[i + 1].inf_inclus = ints[i].inf_inclus

                if ints[i + 1].sup == ints[i].sup:
                    ints[i + 1].sup_inclus = ints[i + 1].sup_inclus or ints[i].sup_inclus
                elif ints[i + 1].sup < ints[i].sup:
                    ints[i + 1].sup = ints[i].sup
                    ints[i + 1].sup_inclus = ints[i].sup_inclus

                ints[i + 1].inf = ints[i].inf
                ints[i] = None
                
            elif ints[i].sup == ints[i + 1].inf and (ints[i].sup_inclus or ints[i + 1].inf_inclus):
                ints[i + 1].inf = ints[i].inf
                ints[i + 1].inf_inclus = ints[i].inf_inclus
                ints[i] = None
                
                
        self.intervalles = [intervalle for intervalle in ints if intervalle is not None]        
        
    def __str__(self):
        return "U".join(str(intervalle) for intervalle in self.intervalles)
        
    def __repr__(self):
        return str(self)
    
    def __nonzero__(self):
        return True
        
    def __neg__(self):
        u"complmentaire"
        return reduce(lambda x, y: x*y, [-intervalle for intervalle in self.intervalles], Intervalle())
        
    def __pos__(self):
        return self
        
    def __sub__(self, y):
        return self*(-y) # "soustraction" peu naturelle ! ;)
        
    def __rsub(self, y):
        return (-self)*y
    
    def __eq__(self, y):
        if len(self.intervalles) <> len(y.intervalles):
            return False
        else:
            test = [i.inf == j.inf and i.sup == j.sup and i.inf_inclus == j.inf_inclus and i.sup_inclus == j.sup_inclus for (i, j) in zip(self.intervalles, y.intervalles)]
            return not (False in test)
        
    def __ne__(self, y):
        return not (self == y)
    
    def __contains__(self, y):
        if isinstance(y, self.__class__) and not isinstance(y, Intervalle):
            return reduce(lambda x, y: x and y, ((intervalle in self) for intervalle in y.intervalles), True)
        return reduce(lambda x, y: x or y, ((y in intervalle) for intervalle in self.intervalles))
    
    
    def extremites(self):
        u"""Retourne les extrmits de chaque intervalle.
        
        Chaque extrmit est donne sous la forme de couples (_float, _str),
        o _str peut prendre les valeurs ".", ")", "(" ou "o".
        Exemple :
        >>> conversion_chaine_ensemble(]-oo;3[U]3;4]U[5;+oo[)
        [(3, 'o'), (4, ')'), (5, '(')]
        """
        
        extremites = []
        for intervalle in self.intervalles:
            if intervalle.inf != -oo:
                if extremites and extremites[-1][0] == intervalle.inf:
                    extremites[-1] = (intervalle.inf, "o")
                else:
                    extremites.append((intervalle.inf, intervalle.inf_inclus and "." or ")"))
            if intervalle.sup != +oo:
                extremites.append((intervalle.sup, intervalle.sup_inclus and "." or "("))
        return extremites
        
    
    def asarray(self, _min, _max, pas):
        u"""Gnre une liste d'objets 'array', correspondant  chaque intervalle.
        
        On se limite  des valeurs comprises entre '_min' et '_max', avec le pas 'pas'."""
        arrays = []
        for intervalle in self.intervalles:
            inf = max(intervalle.inf, _min)
            sup = min(intervalle.sup, _max)
            arrays.append(pylab.append(pylab.arange(float(inf), float(sup), pas), sup))
        return arrays
    
    
    
    
class Vide(Union):
    u"""L'ensemble vide.
    Il n'est pas considr comme un intervalle."""
    def __init__(self):
        self.intervalles = []
    
    def __add__(self, y):
        return y
    
    def __radd__(self, y):
        return y
        
    def __mul__(self, y):
        return self
        
    def __rmul__(self, y):
        return self

    def __neg__(self):
        return Intervalle()
    
    def __str__(self):
        #return ""
        return (u'\xd8').encode(param.encodage) # ""
    
    def __nonzero__(self):
        return False
    
    def __copy__(self):
        return Vide()
    
    def __contains__(self, y):
        return self == y
    
    
    
    
class Intervalle(Union):
    u"""Un intervalle rel non vide.
    Les opration sur les intervalles sont :
    - l'union : symbole "A+B"
    - l'intersection : symbole "A*B"
    - la diffrence : symbole "A-B"
    - le complmentaire : symbole -A
    """
    
    def __init__(self, inf = -sympy.oo, sup = sympy.oo, inf_inclus = True, sup_inclus = True):
        self.inf = inf
        self.sup = sup
        
        if inf == -sympy.oo:
            inf_inclus = False
        if sup == sympy.oo:
            sup_inclus = False
            
        self.inf_inclus = inf_inclus
        self.sup_inclus = sup_inclus
        
        if self.sup < self.inf:
            self.__class__ = Vide
        if self.sup == self.inf:
            if inf_inclus and sup_inclus:
                self.__class__ = Singleton
            else:
                self.__class__ = Vide
        
        if self.__class__ == Vide:
            self.intervalles = []
        else:
            self.intervalles = [self]
     
            
    def __cmp__(self, y):
        c = cmp(self.inf, y.inf)
        if c:   return c
        else:   return cmp(self.sup, y.sup)
     
     
     
        
    def __mul__(self, y):
        u"intersection"
        
        if y == Vide():
            return y
            
        elif isinstance(y, Intervalle): # il s'agit de 2 intervalles
            if self.sup < y.inf or y.sup < self.inf:
                return Vide()
            elif self.sup == y.inf and self.sup_inclus and y.inf_inclus:
                return Singleton(self.sup)
            elif self.inf == y.sup and self.inf_inclus and y.sup_inclus:
                return Singleton(self.inf)
                
            else: # les intervalles se coupent
                if self.inf < y.inf:
                    inf = y.inf
                    inf_inclus = y.inf_inclus
                elif self.inf > y.inf:
                    inf = self.inf
                    inf_inclus = self.inf_inclus                
                else:
                    inf = self.inf
                    inf_inclus = self.inf_inclus and y.inf_inclus
                    
                if y.sup < self.sup:
                    sup = y.sup
                    sup_inclus = y.sup_inclus
                elif y.sup > self.sup:
                    sup = self.sup
                    sup_inclus = self.sup_inclus                
                else:
                    sup = self.sup
                    sup_inclus = self.sup_inclus and y.sup_inclus
                return Intervalle(inf, sup, inf_inclus, sup_inclus)
                
        else: 
            return Union(*[self*i for i in y.intervalles])
    
    
    def __neg__(self):
        return Intervalle(-sympy.oo, self.inf, False, not self.inf_inclus) + Intervalle(self.sup, sympy.oo, not self.sup_inclus)
    
    
    def __str__(self):
        if self.inf == self.sup:
            return "{%s}" %self.inf
        return (self.inf_inclus and "[" or "]") + str(self.inf) + ";" + str(self.sup) + (self.sup_inclus and "]" or "[")
        
    
    def __copy__(self):
        return Intervalle(self.inf, self.sup, self.inf_inclus, self.sup_inclus)
    
    def __contains__(self, y):
        if isinstance(y, self.__class__):
            condition_inf = self.inf < y.inf or (self.inf == y.inf and (self.inf_inclus or not y.inf_inclus))
            condition_sup = self.sup > y.sup or (self.sup == y.sup and (self.sup_inclus or not y.sup_inclus))
            return condition_inf and condition_sup
        elif isinstance(y, Union):
            return reduce(lambda x, y: x and y, ((intervalle in self) for intervalle in y.intervalles), True)
        else:
            return self.inf < y < self.sup or (y == self.inf and self.inf_inclus) or (y == self.sup and self.sup_inclus)
    
    
    
    
        
class Ensemble_discret(Union):
    def __init__(self, *valeurs):
        Union.__init__(self, *[Singleton(val) for val in valeurs])
         
        
        
class Singleton(Intervalle, Ensemble_discret):
    def __init__(self, valeur):
        Intervalle.__init__(self, valeur, valeur)
        
       
    
def formatage_ensemble(chaine):
    u"""Les symboles  utiliser sont 'U' pour l'union, '^' pour l'intersection, '-' ou '\\' pour la soustraction.
    R, R+, R*+, R-, R*- sont aussi accepts."""
    #chaine = chaine.replace("R","Intervalle()")
    chaine = " " + chaine.replace(" ", "") + " " # simplifie l'criture de certains regexp, comme le suivant par exemple

    def f(matchobject): # conversion R -> Intervalle()
        chaine = matchobject.group()
        deb = chaine[0]
        fin = chaine[-1]
        chaine = chaine[1:-1]
        if chaine.endswith("-") and fin in ("[", "]", "{", "(", "R"): # dans "R-{3}", "R-" ne doit pas tre traduit par "]-inf;0]" !
            fin = chaine[-1] + fin
            chaine = chaine[:-1]
        if chaine == "R":
            chaine = "]-oo;oo["
        if chaine == "R*":
            chaine = "(]-oo;oo[-{0})"
        if chaine == "R+":
            chaine = "[0;oo["
        if chaine == "R+*" or chaine == "R*+":
            chaine = "]0;+oo["
        if chaine == "R-":
            chaine = "]-oo;0]"
        if chaine == "R-*" or chaine == "R*-":
            chaine = "]-oo;0["
        return deb + chaine + fin
    n = 0
    while re.search("[^A-TV-Za-z_]R[-+*]?[-+*]?[^A-TV-Za-z_]", chaine):
        chaine = re.sub("[^A-TV-Za-z_]R[-+*]?[-+*]?[^A-TV-Za-z_]", f, chaine) # le U majuscule joue un rle  part (union)
        n += 1
        if n > 1000:
            raise RuntimeError, "Impossible de traduire l'expression %s." %chaine
    
    chaine = chaine.replace("\\", "-")
    chaine = chaine.replace(";[", ";oo[").replace("];", "]oo;")

    def f(matchobject): # conversions ]U[ -> ]+[ , }U[ -> }+[ , etc.
        chaine = matchobject.group()
        return chaine[0] + "+" + chaine[-1]
    chaine = re.sub("[][}{]U[][}{]", f, chaine)

    def f(matchobject): # conversion ]-2;sqrt(3)] -> Intervalle(-2, sqrt(3), False, True)
        chaine = matchobject.group()
        sep = chaine.find(";")
        return "Intervalle(%s,%s,%s,%s)" %(chaine[1:sep],chaine[sep+1:-1],chaine[0]=="[",chaine[-1]=="]")
    chaine = re.sub("[][][-+*/0-9.A-Za-z_)( ]+[;][-+*/0-9.A-Za-z_)( ]+[][]", f, chaine)
    
    def f(matchobject): # conversion {-1;2;sqrt(3)} -> Ensemble_discret(-1, 2, sqrt(3))
        chaine = matchobject.group()
        liste = chaine[1:-1].split(";")
        return "Ensemble_discret(%s)" %",".join(liste)
    chaine = re.sub("[{][-+*/0-9.A-Za-z)(;]+[}]", f, chaine)
    
    def f(matchobject): # conversion inf -> oo
        chaine = matchobject.group()
        return chaine[0] + "oo" + chaine[-1]
    chaine = re.sub("[^A-Za-z]([Ii](nf)?|oo)[^A-Za-z]", f, chaine)
    return chaine
    
    
    
def conversion_chaine_ensemble(chaine):
    chaine = formatage_ensemble(chaine)
    dico = {"__builtins__": None,
#            "Frac": Frac,
            "Intervalle": Intervalle,
            "Ensemble_discret": Ensemble_discret,
            "oo": sympy.oo,
            "False": False,
            "True": True,
            }
    dico.update(math.__dict__)
    return eval(str2(chaine), dico, dico)
